﻿Public Class FilterManager

    Private mInSlot As Int32
    Private mOutSlot As Int32
    Private mSampTime As Int32
    Private mAmplitude As Double = 1
    Private mShift As Double = 0
    Private mInvertFilter As Boolean

    Private mFilterInfo As String

    Private mFirCoeffs(-1) As Double ' FIR coeffs

    Private mCoeffsA(-1) As Double   ' IIR DenomCoeffs
    Private mCoeffsB(-1) As Double   ' IIR NumCoeffs

    Private mCoeffsA0(-1) As Double  ' IIR 2ndOrder coeffs
    Private mCoeffsA1(-1) As Double  ' IIR 2ndOrder coeffs
    Private mCoeffsA2(-1) As Double  ' IIR 2ndOrder coeffs
    Private mCoeffsB0(-1) As Double  ' IIR 2ndOrder coeffs
    Private mCoeffsB1(-1) As Double  ' IIR 2ndOrder coeffs
    Private mCoeffsB2(-1) As Double  ' IIR 2ndOrder coeffs

    Private WithEvents AccTimer As AccurateTimer = New AccurateTimer()
    Private FIR1 As Filter_FIR
    Private IIR1 As Filter_IIR


    Private Enum FilterTypes As Int32
        None
        FIR
        IIR
    End Enum
    Private FilterType As FilterTypes

    Friend Sub SetParams(ByVal inslot As Int32, _
                         ByVal outslot As Int32, _
                         ByVal samptime As Int32, _
                         ByVal amplitude As Double, _
                         ByVal shift As Double, _
                         ByVal invert As Boolean)
        mInSlot = inslot
        mOutSlot = outslot
        mSampTime = samptime
        AccTimer.Interval = mSampTime
        mAmplitude = amplitude
        mShift = shift
        mInvertFilter = invert
    End Sub

    Friend Function GetLenInfo() As String
        Return mFilterInfo
    End Function

    Friend Function GetFilterType() As String
        Return FilterType.ToString
    End Function

    Friend Sub ReadCoefficientFileAndInitFilters()
        AccTimer.Stop()
        ' -------------------------------------
        ReDim mFirCoeffs(-1)
        'ReDim mCoeffsA(-1)
        'ReDim mCoeffsB(-1)
        ReDim mCoeffsA0(-1)
        ReDim mCoeffsA1(-1)
        ReDim mCoeffsA2(-1)
        ReDim mCoeffsB0(-1)
        ReDim mCoeffsB1(-1)
        ReDim mCoeffsB2(-1)
        ' -------------------------------------
        ReadCoeffFile()
        ' -------------------------------------
        Select Case FilterType
            Case FilterTypes.FIR
                FIR1 = New Filter_FIR
                FIR1.Init(mFirCoeffs)
            Case FilterTypes.IIR
                IIR1 = New Filter_IIR
                IIR1.Init(mCoeffsA0, mCoeffsA1, mCoeffsA2, mCoeffsB0, mCoeffsB1, mCoeffsB2)
        End Select
        ' ------------------------------------
        AccTimer.Start()
    End Sub


    Private AnaDigFormat_SectionIndex As Int32

    Private Sub ReadCoeffFile()
        Dim filename As String = LastFilePathAndName
        Dim l As String
        FilterType = FilterTypes.None
        ' -------------------------------------------------------------------------------
        If My.Computer.FileSystem.FileExists(filename) Then
            Dim f As System.IO.StreamReader
            ' ------------------------------------------------ FilterType is FIR or IIR ?
            FilterType = FilterTypes.FIR
            f = IO.File.OpenText(filename)
            For i As Int32 = 1 To 100
                If f.EndOfStream Then Exit For
                l = f.ReadLine().ToLower
                If l.StartsWith("2'nd order") Or l.StartsWith("sect 0") Or _
                                                 l.Contains("digital transfer function coefficients") Then
                    FilterType = FilterTypes.IIR
                    Exit For
                End If
                If l.StartsWith("nth order") Or l.StartsWith("numerator") Then
                    MessageBox.Show("""Nth Order"" IIR Filters are not implemented." + vbCrLf + vbCrLf + _
                                     "The 2'nd order version is equivalent but more robust." + vbCrLf + vbCrLf + _
                                     "Please use a coeff file starting with ""2'nd Order"" or ""Sect 0""", _
                                     "Theremino Filters - Bad coefficient file type", _
                                     MessageBoxButtons.OK, MessageBoxIcon.Error)
                    mFilterInfo = "***ERROR***"
                    FilterType = FilterTypes.None
                    Exit Sub
                End If
            Next
            f.Close()
            ' ------------------------------------------------ read coeffs
            AnaDigFormat_SectionIndex = 0
            Dim validarea As Boolean = True
            f = IO.File.OpenText(filename)
            Do While Not f.EndOfStream
                l = f.ReadLine()
                l = l.Replace("[", "")
                l = l.Replace("]", "")
                l = l.Trim
                If l = "" Then Continue Do
                If l.StartsWith("'") Then Continue Do
                If l.StartsWith("//") Then
                    If l.Contains("Digital Transfer Function Coefficients, k = b0") Then
                        AnaDigFormat_SectionIndex = 1
                    End If
                    Continue Do
                End If
                If l.StartsWith("/*") Then validarea = False : Continue Do
                If l.EndsWith("*/") Then validarea = True : Continue Do
                If l.Contains("define") Then Continue Do
                If l.Contains("{") Then Continue Do
                If l.Contains("}") Then Continue Do
                If validarea Then
                    Select Case FilterType
                        Case FilterTypes.FIR
                            ' ---------------------------------------- read FIR coeffs
                            DecodeCoeffLine_FIR(l)
                        Case FilterTypes.IIR
                            ' ---------------------------------------- read IIR 2ndOrder coeffs
                            l = ReplaceMultipleSpacesAndTrim(l).ToLower
                            If l.StartsWith("nth order") Or l.StartsWith("numerator") Then
                                Exit Do ' stop reading if NthOrder area is detected
                            End If
                            DecodeCoeffLine_IIR(l)
                    End Select
                End If
            Loop
            f.Close()
            ' ------------------------------------------------ final validity test
            Select Case FilterType
                Case FilterTypes.FIR
                    TestIfCoeffsAreAllZero()
                    If mFirCoeffs.Length > 0 Then
                        mFilterInfo = mFirCoeffs.Length.ToString + " taps"
                    Else
                        mFirCoeffs = New Double() {1}
                        mFilterInfo = "*** ERROR ***"
                    End If
                Case FilterTypes.IIR
                    If mCoeffsA0 IsNot Nothing AndAlso mCoeffsA0.Length > 0 Then
                        mFilterInfo = (mCoeffsA0.Length).ToString + " sections"
                    Else
                        mCoeffsA0 = New Double() {1}
                        mCoeffsA1 = New Double() {1}
                        mCoeffsA2 = New Double() {1}
                        mCoeffsB0 = New Double() {1}
                        mCoeffsB1 = New Double() {1}
                        mCoeffsB2 = New Double() {1}
                        mFilterInfo = "*** ERROR ***"
                    End If
            End Select
        Else
            mFilterInfo = "*** ERROR ***"
        End If
    End Sub

    Private Sub TestIfCoeffsAreAllZero()
        ' ------------------------------------------ if saome coeff is not zero return
        For i As Int32 = 0 To mFirCoeffs.Length - 1
            If mFirCoeffs(i) <> 0 Then Return
        Next
        ' ------------------------------------------ else clear the array
        ReDim mFirCoeffs(-1)
    End Sub

    Private Sub DecodeCoeffLine_FIR(ByVal l As String)
        ReDim Preserve mFirCoeffs(mFirCoeffs.Length)
        mFirCoeffs(mFirCoeffs.Length - 1) = Val(l.Replace(",", "."))
    End Sub

    Private Sub DecodeCoeffLine_IIR(ByVal l As String)
        l = l.Replace("=", "")
        l = ReplaceMultipleSpacesAndTrim(l)
        Dim sa As String() = l.Split(" "c)
        Static n As Int32
        If sa.Length = 2 Then
            Dim v As Double = Val(sa(1).Replace(",", "."))
            Dim sa0 As String() = sa(0).Split(","c)
            If sa0.Length = 1 Then
                ' ------------------------------------------------------------------ IowaHills format
                Select Case sa(0)
                    Case "sect" : n = CInt(Val(sa(1)))
                    Case "a0" : UpdateCoefficient(mCoeffsA0, n, v)
                    Case "a1" : UpdateCoefficient(mCoeffsA1, n, v)
                    Case "a2" : UpdateCoefficient(mCoeffsA2, n, v)
                    Case "b0" : UpdateCoefficient(mCoeffsB0, n, v)
                    Case "b1" : UpdateCoefficient(mCoeffsB1, n, v)
                    Case "b2" : UpdateCoefficient(mCoeffsB2, n, v)
                End Select
            End If
            If sa0.Length = 2 Then
                n = (CInt(Val(sa0(1))) - 1) * 2 + AnaDigFormat_SectionIndex
                Select Case sa0(0)
                    ' -------------------------------------------------------------- AnaDig format
                    Case "a(0" : UpdateCoefficient(mCoeffsA0, n, v)
                    Case "a(1" : UpdateCoefficient(mCoeffsA1, n, v)
                    Case "a(2" : UpdateCoefficient(mCoeffsA2, n, v)
                    Case "b(0" : UpdateCoefficient(mCoeffsB0, n, v)
                    Case "b(1" : UpdateCoefficient(mCoeffsB1, n, v)
                    Case "b(2" : UpdateCoefficient(mCoeffsB2, n, v)
                End Select
            End If
        End If
    End Sub

    Private Sub UpdateCoefficient(ByRef array As Double(), ByVal index As Int32, ByVal value As Double)
        If index >= array.Length Then
            ReDim Preserve array(index)
        End If
        array(index) = value
    End Sub


    ' ==============================================================
    '  REPLACE MULTIPLE SPACES and TABS and TRIM
    ' ============================================================== 
    Friend Function ReplaceMultipleSpacesAndTrim(ByVal s As String) As String
        s = s.Replace(vbTab, " ")
        While s.Contains("  ")
            s = s.Replace("  ", " ")
        End While
        Return s.Trim
    End Function

    Private Sub AccTimer_TimerTick(ByVal milliseconds As Double) Handles AccTimer.Tick
        Dim n As Double = Slots.ReadSlot(mInSlot)
        Dim n2 As Double
        Select Case FilterType
            Case FilterTypes.FIR : n2 = FIR1.NewSample(n)
            Case FilterTypes.IIR : n2 = IIR1.NewSample_2ndOrderForm1(n)
        End Select
        If mInvertFilter Then n2 = n - n2
        Slots.WriteSlot(mOutSlot, n2 * mAmplitude + mShift)
    End Sub

    ' ---------------------------------------------
    '  IIR Methods ordered by overflow resistance
    ' ---------------------------------------------
    ' 1 - 2ndOrderF1
    ' 2 - 2ndOrderF2
    ' 3 - NthOrderF2
    ' 4 - NthOrderF1B
    ' 5 - NthOrderF1A

End Class
